# -*- coding: utf-8 -*-
"""
The :py:mod:`tdda.constraints` module includes a set of examples.

To copy these examples to your own *constraints-examples* subdirectory
(or to a location of your choice), run the command::

    tdda examples constraints [mydirectory]

"""
from __future__ import absolute_import

from tdda import examples

if __name__ == '__main__':
    examples.copy_main('constraints')

